//FEM code to solve the Lid Driven Cavity Stokes problem. Created by Scott Ladenheim, 4/10/2014

/*
 *  These are all of the include files.
* Include files tell your program what set of files to include when it runs.
* These are the .h files and differnt files are found in different sub directories.
* For instance, deal.II/base/ or deal.II/lac
*
* The last set of includes are standard c++ libraries which give you access to
* various input/output classes
* vector classes
* etc.
*/
#include <deal.II/base/quadrature_lib.h>
#include <deal.II/base/function.h>
#include <deal.II/base/logstream.h>
#include <deal.II/base/utilities.h>
#include <deal.II/base/convergence_table.h>
#include <deal.II/base/timer.h>

#include <deal.II/lac/vector.h>
#include <deal.II/lac/full_matrix.h>
#include <deal.II/lac/sparse_matrix.h>
#include <deal.II/lac/sparse_direct.h>
#include <deal.II/lac/constraint_matrix.h>
#include <deal.II/lac/solver_gmres.h>
#include <deal.II/lac/solver_minres.h>
#include <deal.II/lac/precondition.h>
#include <deal.II/lac/solver_control.h>
#include <deal.II/lac/block_vector.h>
#include <deal.II/lac/block_sparse_matrix.h>
#include <deal.II/lac/block_sparsity_pattern.h>
#include <deal.II/lac/iterative_inverse.h>

#include <deal.II/grid/tria.h>
#include <deal.II/grid/grid_generator.h>
#include <deal.II/grid/tria_accessor.h>
#include <deal.II/grid/tria_iterator.h>
#include <deal.II/grid/tria_boundary_lib.h>
#include <deal.II/grid/grid_out.h>

#include <deal.II/dofs/dof_handler.h>
#include <deal.II/dofs/dof_accessor.h>
#include <deal.II/dofs/dof_tools.h>
#include <deal.II/dofs/dof_renumbering.h>

#include <deal.II/fe/fe_q.h>
#include <deal.II/fe/fe_system.h>
#include <deal.II/fe/fe_values.h>
#include <deal.II/fe/component_mask.h>

#include <deal.II/numerics/vector_tools.h>
#include <deal.II/numerics/matrix_tools.h>
#include <deal.II/numerics/data_out.h>

#include <typeinfo>
#include <fstream>
#include <ostream>
#include <iterator>
#include <vector>
#include <sstream>
#include <algorithm>
#include <iostream>
#include <iomanip>
#include <cmath>

namespace StokesProblem
{
using namespace dealii;

/* Here we declare a class called Stokes
 * The stokes class has a public constructor and
 * public function called run.
 *
 * It has private functions called
 * make_grid
 * setup_system
 * assemble_system
 * solve
 * output_results
 *
 * It has private member variables for the
 * degree of the FE approximation
 * triangulation object
 * fe system
 * degree of freedom handler (dofhandler)
 * constraint matrix
 * sparsity_pattern
 * system_matrix
 * solution
 * system_rhs
 *
 *
 */
template<int dim>
class stokes
{
public:
	stokes(const unsigned int degree,const double viscosity);
	void run();

private:
	void make_grid();
	void setup_system();
	void assemble_system();
	void solve(const unsigned int cycle);
	void output_results(const unsigned int cycle);

	const unsigned int degree;

	Triangulation<dim>   triangulation;
	FESystem<dim>        fe;
	DoFHandler<dim>      dof_handler;

	ConstraintMatrix     constraints;

	BlockSparsityPattern      sparsity_pattern;
	BlockSparseMatrix<double> system_matrix;

	BlockVector<double> solution;
	BlockVector<double> system_rhs;

	const double nu;
};

/*
 * This function defines the righthand side of the stokes problem.
 * This function encodes the vector right hand-side and is defined on the (u,p) space.
 * This means that we give a vector f(x) and for incompressible problems, g(x)=0
 */

template <int dim>
class RightHandSide : public Function<dim>
{
public:
	RightHandSide () : Function<dim>(dim+1) {}


	virtual void vector_value (const Point<dim>   &p,
			Vector<double> &values) const;

	virtual void vector_value_list (const std::vector<Point<dim> > &points,
			std::vector <Vector<double> >   &value_list) const;

};

template <int dim>
inline
void RightHandSide<dim>::vector_value (const Point<dim> &p,
		Vector<double>   &values) const
{
	values(0)= 0.0;
	values(1)= 0.0;
	values(2)= 0.0;

}//RightHandside<dim>::vector_value

template <int dim>
void RightHandSide<dim>::vector_value_list (const std::vector<Point<dim> > &points,
		std::vector<Vector<double> >   &value_list) const
{
	const unsigned int n_points = points.size();

	for (unsigned int p=0; p<n_points; ++p)
		RightHandSide<dim>::vector_value (points[p],
				value_list[p]);
}

template <int dim>
class StokesBoundaryValues : public Function<dim>
{
public:
	StokesBoundaryValues() : Function<dim>(dim+1) {}
	virtual double value (const Point<dim> &p, const unsigned int component =0) const;
	virtual void vector_value(const Point<dim> &p, Vector<double> &value) const;

};

template <int dim>
double
StokesBoundaryValues<dim>::value (const Point<dim>  &p,
                            const unsigned int component) const
{
	// Lid Driven Cavity Boundary Conditions
	if (component == 0)
		return ( ( p[dim-1]==1) ? 1 : 0 );
	else if (component == dim-1)
		return 0.0 ;
	else
		return 0.0;
}

template <int dim>
void
StokesBoundaryValues<dim>::vector_value (const Point<dim> &p,
                                   Vector<double>   &values) const
{
  for (unsigned int c=0; c<this->n_components; ++c)
    values(c) = StokesBoundaryValues<dim>::value (p, c);
}

template <class Matrix>
class InverseMatrix : public Subscriptor
{
public:
	InverseMatrix (const Matrix &m);
	void vmult (Vector<double> &dst, const Vector<double> &src) const;

private:
	const SmartPointer<const Matrix> matrix;
};

template <class Matrix>
InverseMatrix<Matrix>::InverseMatrix(const Matrix &m)
:
matrix (&m)
{}

template <class Matrix>
void InverseMatrix<Matrix>::vmult(Vector<double> &dst, const Vector<double> &src) const
{
	dst=0.0;
	SparseDirectUMFPACK direct;
	direct.initialize (*matrix);
	direct.vmult (dst, src);
};

template<int dim>
stokes<dim>::stokes (const unsigned int degree, const double viscosity)
:
degree(degree),
triangulation(Triangulation<dim>::limit_level_difference_at_vertices),
fe(FE_Q<dim>(degree),dim,FE_Q<dim>(degree),1),
dof_handler(triangulation),
nu(viscosity)
{}

template<int dim>
void stokes<dim>::make_grid()
{
	GridGenerator::hyper_cube(triangulation,0.,1.);

	triangulation.refine_global(2); // h=2^-2 =1/4 is starting mesh width
}

template<int dim>
void stokes<dim>::setup_system()
{
	system_matrix.clear ();

	dof_handler.distribute_dofs (fe);

	std::vector<unsigned int> block_component (dim+1,0);
	block_component[dim] = 1;
	DoFRenumbering::component_wise (dof_handler, block_component);

	{
		constraints.clear ();
		FEValuesExtractors::Vector velocities(0);

		DoFTools::make_hanging_node_constraints (dof_handler,
				constraints);

		VectorTools::interpolate_boundary_values (dof_handler,
				0,
				StokesBoundaryValues<dim>(),
				constraints,
				fe.component_mask(velocities));
	}

	FEValuesExtractors::Scalar pressure(dim);
	ComponentMask pressure_mask = fe.component_mask (pressure);

	//Boundary constraint
	std::vector<bool> boundary_dofs (dof_handler.n_dofs(), false);
	DoFTools::extract_boundary_dofs (dof_handler, pressure_mask,boundary_dofs);

	const unsigned int first_boundary_dof
	= std::distance (boundary_dofs.begin(), std::find (boundary_dofs.begin(),
			boundary_dofs.end(), true));

	constraints.add_line (first_boundary_dof);
	for (unsigned int i=first_boundary_dof+1; i<dof_handler.n_dofs(); ++i)
		if (boundary_dofs[i] == true)
			constraints.add_entry (first_boundary_dof,
					i, -1);

	constraints.close ();

	std::vector<types::global_dof_index> dofs_per_block (2);
	DoFTools::count_dofs_per_block (dof_handler, dofs_per_block, block_component);
	const unsigned int n_u = dofs_per_block[0],
			n_p = dofs_per_block[1];

	std::cout << "   Number of active cells: "
			<< triangulation.n_active_cells()
			<< "\n"
			<< "   Number of degrees of freedom: "
			<< dof_handler.n_dofs()
			<< " (" << n_u << '+' << n_p << ')'
			<< "\n";
	{
		BlockCompressedSimpleSparsityPattern csp (2,2);
		csp.block(0,0).reinit (n_u, n_u);
		csp.block(1,0).reinit (n_p, n_u);
		csp.block(0,1).reinit (n_u, n_p);
		csp.block(1,1).reinit (n_p, n_p);
		csp.collect_sizes();

		DoFTools::make_sparsity_pattern (dof_handler, csp, constraints, false);

		constraints.condense(csp);

		sparsity_pattern.copy_from (csp);
	}
	system_matrix.reinit (sparsity_pattern);

	solution.reinit (2);
	solution.block(0).reinit (n_u);
	solution.block(1).reinit (n_p);
	solution.collect_sizes ();

	system_rhs.reinit (2);
	system_rhs.block(0).reinit (n_u);
	system_rhs.block(1).reinit (n_p);
	system_rhs.collect_sizes ();
}

template <int dim>
void stokes<dim>::assemble_system ()
{
	system_matrix=0;
	system_rhs=0;

	QGauss<dim>   quadrature_formula(degree+2);
	FEValues<dim> fe_values (fe, quadrature_formula,
			update_values    |
			update_quadrature_points  |
			update_JxW_values |
			update_gradients);

	const unsigned int   dofs_per_cell   = fe.dofs_per_cell;
	const unsigned int   n_q_points      = quadrature_formula.size();

	FullMatrix<double>   local_matrix (dofs_per_cell, dofs_per_cell);
	Vector<double>       local_rhs (dofs_per_cell);

	std::vector<types::global_dof_index> local_dof_indices (dofs_per_cell);

	const RightHandSide<dim>          right_hand_side;
	std::vector<Vector<double> >      rhs_values (n_q_points, Vector<double>(dim+1));

	const FEValuesExtractors::Vector velocities (0);
	const FEValuesExtractors::Scalar pressure (dim);

	std::vector<SymmetricTensor<2,dim> > symgrad_phi_u (dofs_per_cell);
	std::vector<double>                  div_phi_u   (dofs_per_cell);
	std::vector<double>                  phi_p       (dofs_per_cell);

	typename DoFHandler<dim>::active_cell_iterator
	cell = dof_handler.begin_active(),
	endc = dof_handler.end();
	for (; cell!=endc; ++cell)
	{
		fe_values.reinit (cell);
		local_matrix = 0;
		local_rhs = 0;
		right_hand_side.vector_value_list(fe_values.get_quadrature_points(),
				rhs_values);
		for (unsigned int q=0; q<n_q_points; ++q)
		{
			for (unsigned int k=0; k<dofs_per_cell; ++k)
			{
				symgrad_phi_u[k] = fe_values[velocities].symmetric_gradient (k, q);
				div_phi_u[k]     = fe_values[velocities].divergence (k, q);
				phi_p[k]         = fe_values[pressure].value (k, q);
			}
			for (unsigned int i=0; i<dofs_per_cell; ++i)
			{
				for (unsigned int j=0; j<=i; ++j)
				{
					local_matrix(i,j) += (2*symgrad_phi_u[i] * symgrad_phi_u[j]- div_phi_u[i] * phi_p[j]
					 - phi_p[i] * div_phi_u[j]) * fe_values.JxW(q);
				}
				const unsigned int component_i =
						fe.system_to_component_index(i).first;
				local_rhs(i) += fe_values.shape_value(i,q) *
						rhs_values[q](component_i) *
						fe_values.JxW(q);
			}
		}
		for (unsigned int i=0; i<dofs_per_cell; ++i)
			for (unsigned int j=i+1; j<dofs_per_cell; ++j)
				local_matrix(i,j) = local_matrix(j,i);
		cell->get_dof_indices (local_dof_indices);
		constraints.distribute_local_to_global (local_matrix, local_rhs,
				local_dof_indices,
				system_matrix, system_rhs);
	}
}

template<int dim>
void stokes<dim>::solve(const unsigned int cycle)
{
	SparseDirectUMFPACK direct_solver;
	direct_solver.initialize (system_matrix);
	direct_solver.vmult (solution, system_rhs);

	constraints.distribute(solution);

	system_matrix.clear();
}

template<int dim>
void stokes<dim>::output_results(const unsigned int cycle)
{
	std::vector<std::string> solution_names(dim,"velocity");
	solution_names.push_back("pressure");

	std::vector<DataComponentInterpretation::DataComponentInterpretation>
	data_component_interpretation
	(dim, DataComponentInterpretation::component_is_part_of_vector);
	data_component_interpretation
	.push_back (DataComponentInterpretation::component_is_scalar);

	DataOut<dim> data_out;
	data_out.attach_dof_handler (dof_handler);
	data_out.add_data_vector (solution, solution_names,
			DataOut<dim>::type_dof_data,
			data_component_interpretation);
	data_out.build_patches ();

	std::ostringstream sol_name;
	sol_name << "lid_driven_cavity_solution" << dim << "-cycle" << Utilities::int_to_string(cycle,1) <<".vtk";

	std::ofstream output (sol_name.str().c_str());
	data_out.write_vtk (output);
}

template<int dim>
void stokes<dim>::run ()
{
	std::cout << "Solving problem in " << dim << "dimensions"<<  std::endl;

	 for (unsigned int cycle=0; cycle<6; ++cycle)
	 {
		 std::cout << "Cycle " << cycle << ':' << std::endl;

		 if (cycle == 0)
		 {
			 std::cout << "Making grid..." << std::endl;
			 make_grid();
		 }
		 else
			 triangulation.refine_global(1);

		 std::cout << "Setting up system ..." << std::endl;
		 setup_system ();

		 std::cout << "Assembling..." << std::endl;
		 assemble_system ();

		 std::cout << "Solving..." << std::endl;
		 solve (cycle);

		 std::cout << "Output results..." << std::endl;
		 output_results (cycle);
	 }
}
//end of StokesProblem namespace
}

int main ()
{
	using namespace dealii;
	using namespace StokesProblem;

	deallog.depth_console (0);

	{
		stokes<2> stokes_2d(1,1.0);
		stokes_2d.run();
	}

	return 0;
}
